@@ -6,7 +6,7 @@ import sys
 from asyncio.subprocess import PIPE, create_subprocess_exec
 from pathlib import Path
 from types import ModuleType
-from typing import Any, Iterable
+from typing import Any, Iterable, Union
 
 log = logging.getLogger(__name__)
 
@@ -25,89 +25,86 @@ def _listify(value):
     return value if _nonstriterable(value) else [value]
 
 
+def _convert_kwargs(kwargs):
+    """Convert arguments passed by keywords into those that are passed as
+    arguments to the subprocess vs those that are passed to the subprocess.
+
+    Arguments that start with underscore are treated specially and are
+    passed through to the underlying subprocess call.
+
+    All other arguments are passed through as arguments to the program
+    opened in the subprocess call, according to the following rules:
+
+    - underscores are converted to dashes: shell convention tends towards
+        dashes, while tokens in Python can only contain underscores.
+    - if an argument value is True (eg. foo=True), it is passed through to
+        the underlying subprocess call as either -f or --foo, depending on
+        how many letters the argument has.
+    - if _env is passed, is it merged with the wider os environment and
+        passed to the subprocess's env call
+    """
+    subprocess_args = []
+    subprocess_kwargs = {}
+    for k, vs in kwargs.items():
+        if k.startswith('_'):
+            subprocess_kwargs[k[1:]] = vs  # pass arg (- underscore) through to 'run'
+            continue
+
+        key = f"{'-' * min(len(k), 2)}{k.replace('_', '-')}"
+        for v in _listify(vs):
+            if v is True:  # cmd(foo=True) -> cmd --foo
+                subprocess_args.append(key)
+            else:
+                subprocess_args.extend([key, str(v)])
+
+    return subprocess_args, subprocess_kwargs
+
+
 class Command:
-    _default_kwargs = {'stdout': PIPE, 'stderr': PIPE}
+    _default_kwargs = {'_stdout': PIPE, '_stderr': PIPE}
     _command: list[str]
     _kwargs: dict[str, Any]
     _env: dict[str, str]
     _check: bool
-    def __init__(self, name, *args, _check=True, _env={}, **kwargs):
-        self._command = [name, *args]
-        self._kwargs = self._default_kwargs | kwargs
-        # environment variables need to be treated specially:
-        # - they need to be merged when you refine commands
-        # - they should automatically inherit from os.environ when called
-        self._env = {} | _env
+    def __init__(self, *args, _check=True, **kwargs):
+        if not args:
+            raise Exception("Must provide command name")
+
+        self._name = args[0].replace('_', '-')
         self._check = _check
+        self._kwargs = kwargs
+        subprocess_args, subprocess_kwargs = _convert_kwargs(self._default_kwargs | kwargs)
+        self._command = [*args] + subprocess_args
+        self._subprocess_kwargs = subprocess_kwargs
 
     def _bake(self, *args, **kwargs):
-        check = kwargs.pop('_check', self._check)
-        converted_args, converted_kwargs = self._convert_kwargs(kwargs)
-        cmd = self._command + [*args] + converted_args
-        kwargs = self._kwargs | converted_kwargs
-        return Command(*cmd, _check=check, _env=self._env, **kwargs)
+        return Command(*self._command, *args, **(self._kwargs | kwargs))
 
     def __call__(self, *args, **kwargs):
+        kwargs = kwargs.copy()
+        if env := kwargs.get('_env'):
+            # merge provided env with os.environ
+            kwargs['_env'] = os.environ | {k: str(v) for k, v in env.items()}
+
         cmd = self._bake(*args, **kwargs)
-        log.warning(f"Executing: {cmd}")
-        echo = logging.root.level <= logging.INFO
-        return Result(cmd, echo)
+        return Result(cmd, logging.root.level <= logging.INFO)
 
     def __getitem__(self, key):
         new_cmd = self._command + list([key] if isinstance(key, str) else key)
-        return Command(*new_cmd, _check=self._check, _env=self._env, **self._kwargs)
+
+        return Command(*new_cmd, _check=self._check, **self._kwargs)
 
     def __getattr__(self, name):
-        return self[name.replace('_', '-')]
+        return self[name]
 
     def __str__(self):
         return str(self._command)
 
     def __repr__(self):
-        return f"{self.__class__.__name__}({self._command!r})"
+        return f"{self.__class__.__name__}({self._command})"
 
     def __or__(self, other):
-        return self() | other
-
-    def _convert_kwargs(self, kwargs):
-        """Convert arguments passed by keywords into those that are passed as
-        arguments to the subprocess vs those that are passed to the subprocess.
-
-        Arguments that start with underscore are treated specially and are
-        passed through to the underlying subprocess call.
-
-        All other arguments are passed through as arguments to the program
-        opened in the subprocess call, according to the following rules:
-
-        - underscores are converted to dashes: shell convention tends towards
-          dashes, while tokens in Python can only contain underscores.
-        - if an argument value is True (eg. foo=True), it is passed through to
-          the underlying subprocess call as either -f or --foo, depending on
-          how many letters the argument has.
-        - if _env is passed, is it merged with the wider os environment and
-          passed to the subprocess's env call
-        """
-        converted_args = []
-        converted_kwargs = {}
-        for k, vs in kwargs.items():
-            if k.startswith('_'):
-                converted_kwargs[k[1:]] = vs  # pass arg (- underscore) through to 'run'
-                continue
-
-            key = f"{'-' * min(len(k), 2)}{k.replace('_', '-')}"
-            for v in _listify(vs):
-                if v is True:  # cmd(foo=True) -> cmd --foo
-                    converted_args.append(key)
-                else:
-                    converted_args.extend([key, str(v)])
-
-        env = converted_kwargs.get('env', {})
-        if self._env or env:
-            converted_kwargs['env'] = os.environ | {
-                k: str(v) for k, v in (self._env | env).items()
-            }
-
-        return converted_args, converted_kwargs
+        return Pipeline(self, other)
 
 
 async def _read(buffer, stream, echo=True, color=None):
@@ -124,16 +121,24 @@ async def _read(buffer, stream, echo=True, color=None):
     return buffer.getvalue()
 
 
+async def _run(command: Command):
+    """Run a Command, return the Process object"""
+    cmd = list(map(str, command._command))
+    log.warning(f"Executing: {cmd}")
+    return await create_subprocess_exec(*cmd, **command._subprocess_kwargs)
+
+
 class Result:
     def __init__(self, command: Command, echo=True):
         self._command = command
         self._stdout = io.BytesIO()
         self._stderr = io.BytesIO()
 
-        process_coroutine = create_subprocess_exec(*command._command, **command._kwargs)
-        self._process = LOOP.run_until_complete(process_coroutine)
-        LOOP.create_task(_read(self._stdout, self._process.stdout, echo))
-        LOOP.create_task(_read(self._stderr, self._process.stderr, echo, color=STDERR_COLOR))
+        self._process = LOOP.run_until_complete(_run(command))
+        LOOP.run_until_complete(asyncio.gather(
+            _read(self._stdout, self._process.stdout, echo),
+            _read(self._stderr, self._process.stderr, echo, color=STDERR_COLOR),
+        ))
 
         if self._command._check and self.code:
             # lazily evaluates code, which blocks, only if _check
@@ -174,6 +179,9 @@ class Result:
     def __str__(self):
         return self.stdout.decode().strip()
 
+    def __bytes__(self):
+        return self.stdout
+
     def __iter__(self):
         yield from str(self).splitlines()
 
@@ -183,6 +191,9 @@ class Result:
             f.write(getattr(self, stream))
         return self
 
+    # todo: make these redirection writes async, but ensure any read tasks
+    # finish when the process closes.
+    # Probaxbly need to put all reads/writes on a queue
     def __gt__(self, other):
         return self._write(Path(other), 'stdout', 'wb')
 
@@ -195,13 +206,18 @@ class Result:
     def __pow__(self, other):
         return self._write(Path(other), 'stderr', 'ab')
 
-    def __or__(self, other):
-        """Pipe two Commands together
 
-        This directly connects the stdout left -> right like a shell pipeline
-        so that commands run in parallel
-        """
-        return other._bake(_stdin=self._stdout)
+class Pipeline:
+    def __init__(self, left: Union[str, Result, Command, "Pipeline"], right: Command):
+        self.left = left
+        self.right = right
+
+    def __call__(self):
+        if isinstance(self.left, Command):
+            result = self.left(_check=False)
+            return self.right(_stdin=result._process.stdin)
+
+        self.right._bake(_stdin=self.left.stdout)
 
 
 class _AushModule(ModuleType):
@@ -214,7 +230,7 @@ class _AushModule(ModuleType):
             # doesn't change the working directory of this process
             return os.chdir
 
-        return Command(name.replace('_', '-'))
+        return Command(name)
 
     __getattr__ = __getitem__
 
